#ifndef BOOST_STACKTRACE_WIN32_HPP
#define BOOST_STACKTRACE_WIN32_HPP

#include <windows.h>

// <dbghelp.h> has conflicts with <imagehlp.h>, so it cannot be #include'd
// <imagehlp.h> contains everything we would need from <dbghelp.h> anyway
#include <imagehlp.h>
#include <tlhelp32.h>
#include <psapi.h>

#include <cstring>
#include <string>
#include <sstream>
#include <iomanip>
#include <algorithm>

#include <boost/lexical_cast.hpp>
#include <boost/utility.hpp>
#include <boost/smart_ptr.hpp>

//
// Support for Borland intrinsic functions
//  (this should go in Boost.Config)
//

#if defined(__TURBOC__) && defined(strcpy)
  namespace std {
    // Note: we cannot just do a "using ::strcpy;" here because BCB6 will choke on it
    char * __strcpy__(char * const dest, const char * const src)
    { return ::__strcpy__(dest, src); }
  } // namespace std
#endif

namespace boost {

namespace details {
namespace stack {

//
// Win32/64 helper macros (defined by system header files or the compiler):
//  _WIN64 64-bit
//  _WIN32 64-bit or 32-bit
//
//  _M_IX86 Intel 32-bit
//  _M_IA64 Intel 64-bit
//  _M_AMD64 AMD 64-bit
//

// Occasionally, a reinterpret_cast is used in these functions.  These casts
//  are safe; just keep in mind the following:
//   void * - 32 bits on Win32, 64 bits on Win64
//   DWORD64 - 64 bits on Win32, 64 bits on Win64
//   DWORD - 32 bits on Win32, unspecified on Win64 (probably 32 bits)
//   SIZE_T - 32 bits on Win32, and 64 bits on Win64 (unsigned)

//
// General string manipulation functions
//

// Removes whitespace (" \r\n\t") from the beginning and end of a string
std::string trim(const std::string & x)
{
  const std::string::size_type begin = x.find_first_not_of(" \r\n\t");
  if (begin == std::string::npos)
    return std::string();
  return x.substr(begin, x.find_last_not_of(" \r\n\t") - begin + 1);
}

// Converts 'value' to a hex string, prefixed with "0x"
// This function is intended for use only with builtin types!
template <typename T>
std::string to_hex_string(const T & value)
{
  std::ostringstream out;
  out << "0x" << std::hex << std::setfill('0') << std::setw(sizeof(T) * 2) << value;
  return out.str();
}

// Replaces all line ends ("\r\n") with spaces (" ")
struct is_cr_or_nl
{
  bool operator()(const char x) const
  { return (x == '\r' || x == '\n'); }
};
std::string flatten(std::string x)
{
  std::replace_if(x.begin(), x.end(), is_cr_or_nl(), ' ');
  return x;
}

//
// Debugging functions
//
//
// The stack tracing implementation for Win32 will report any failed functions
//  to OutputDebugString, prefixed by the string "Boost.StackWalk: ".
//

// This function adds the common prefix, removes any embedded newlines and
//  carriage returns, and sends it to OutputDebugString
void debug(const std::string & str)
{
  OutputDebugString(("Boost.StackWalk: " + flatten(str)).c_str());
}

// Retrieves the current Win32 error and also clears it
DWORD debug_Win32_get_error()
{
  const DWORD ret = GetLastError();
  SetLastError(0);
  return ret;
}

// Formats a Win32 error message from an error code; whitespace at the end of
//  the message is trimmed off, but other whitespace may be embedded within the
//  returned string
std::string debug_Win32_format(const DWORD code = debug_Win32_get_error())
{
  void * lpMsgBuf;

  if (!FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
      FORMAT_MESSAGE_IGNORE_INSERTS,
      0, code, 0, (char *) &lpMsgBuf, 0, 0))
    return "Unknown error " + lexical_cast<std::string>(code);

  std::string ret = (char *) lpMsgBuf;

  ret = trim(ret);

  if (LocalFree(lpMsgBuf) != 0)
    ret += " <LocalFree for formatting Win32 error message failed>";

  return ret;
}

// Sends out a complete Win32-based-error message of the form:
//    Boost.StackTrace: Win32 error `$1' from `$2'
//  where $1 is the Win32 formatted error message and $2 is the function that
//  caused the error.
void debug_Win32(const std::string & funcname,
    const DWORD code = debug_Win32_get_error())
{
  debug("Win32 error `" + debug_Win32_format(code) + "' from `" + funcname + "'");
}

//
// General helpers
//

// This class provides thread-safe singleton construction and destruction
template <typename T>
class singleton
{
  private:
    static LONG flag;
    static T * data;

    // No objects of this type ever exist
    singleton();

  public:
    // Calling this function will return a global object of type T.
    // The first time this function is called, it will construct the global
    //   object in a thread-safe fashion.
    // The object has static duration, so it will exist until the end of the
    //   program.
    // Note: Objects with static duration are destroyed in the reverse order
    //   of their construction.
    static T & instance()
    {
      if (!data)
        if (InterlockedIncrement(&flag) == 0)
        {
          // Only one thread (the first thread) will ever execute this code
          static scoped_ptr<T> deleter(new T);
          data = deleter.get();
        }
        else
        {
          T * volatile * const pdata = &data;
          // Busy-wait; volatile protects against compiler optimization
          while (!*pdata)
            Sleep(0);
          // Have to return through the volatile pointer in this case
          return **pdata;
        }
      return *data;
    }
};
// These are statically initialized (guaranteed to be done before dynamic
//   initialization); this allows dynamic initialization code to call
//   singleton::instance().
// See the Win32 docs for InterlockedIncrement to see why -1 is used.
template <typename T>
LONG singleton<T>::flag = -1;
template <typename T>
T * singleton<T>::data = 0;

// Simple class that closes a handle on scope exit
class file_handle: noncopyable
{
  private:
    HANDLE handle_;

  public:
    explicit file_handle(const HANDLE nhandle)
    :handle_(nhandle) { }

    ~file_handle()
    {
      if (!CloseHandle(handle_))
        debug_Win32("CloseHandle");
    }

    HANDLE handle() const { return handle_; }
};

// Yet Another Critical Section Class
class critical_section: noncopyable
{
  private:
    CRITICAL_SECTION sec_;

  public:
    critical_section() { InitializeCriticalSection(&sec_); }
    explicit critical_section(const DWORD spin_count) { InitializeCriticalSectionAndSpinCount(&sec_, spin_count); }

    ~critical_section() { DeleteCriticalSection(&sec_); }

    void lock() { EnterCriticalSection(&sec_); }
    bool try_lock() { return TryEnterCriticalSection(&sec_); }
    void unlock() { LeaveCriticalSection(&sec_); }

    class scoped_lock: noncopyable
    {
      private:
        critical_section & s_;

      public:
        explicit scoped_lock(critical_section & ns):s_(ns) { s_.lock(); }

        ~scoped_lock() { s_.unlock(); }
    };
};

//
// "Global" variables
//

template <typename T>
struct singleton_tag
{
  typedef typename T::data_type data_type;
  data_type data;

  singleton_tag()
  :data(T::default_value()) { }
};

struct dbghelp_dll_location_tag
{
  typedef std::string data_type;
  static data_type default_value() { return "dbghelp.dll"; }
};
struct toolhelp_dll_location_tag
{
  typedef std::string data_type;
  static data_type default_value() { return "kernel32.dll"; }
};
struct psapi_dll_location_tag
{
  typedef std::string data_type;
  static data_type default_value() { return "psapi.dll"; }
};
struct symbol_search_path_tag
{
  typedef std::string data_type;
  static data_type default_value() { return data_type(); }
};

typedef singleton<singleton_tag<dbghelp_dll_location_tag> > dbghelp_dll_location;
typedef singleton<singleton_tag<toolhelp_dll_location_tag> > toolhelp_dll_location;
typedef singleton<singleton_tag<psapi_dll_location_tag> > psapi_dll_location;
typedef singleton<singleton_tag<symbol_search_path_tag> > symbol_search_path;
typedef singleton<critical_section> dbghelp_lock;

//
// DLL functions and classes
//

// This is a class that represents a DLL that has been loaded into memory
class dll_base: noncopyable
{
  private:
    HMODULE module_;

  protected:
    // Returns the address of the specified procedure, or 0 if not found.
    void * load_proc(const char * const name)
    {
      void * const ret = GetProcAddress(module_, name);
      if (ret == 0)
        debug_Win32("GetProcAddress (looking up `" + std::string(name) + "')");
      return ret;
    }

  public:
    // Loads the named DLL into memory
    //  If the DLL is not successfully loaded, module() will return 0.
    explicit dll_base(const std::string & name)
    {
      module_ = LoadLibrary(name.c_str());
      if (module_ == 0)
        debug_Win32("LoadLibrary (loading `" + name + "')");
    }

    HMODULE module() const { return module_; }

    // Unloads the DLL from memory if it was successfully loaded
    ~dll_base()
    {
      if (module_ != 0)
        if (!FreeLibrary(module_))
          debug_Win32("FreeLibrary");
    }
};

// This class represents kernel32.dll and the toolhelp functions from it
struct toolhelp_dll: public dll_base
{
  typedef HANDLE (__stdcall * CreateToolhelp32Snapshot_type)(DWORD, DWORD);
  typedef BOOL (__stdcall * Module32First_type)(HANDLE, LPMODULEENTRY32);
  typedef BOOL (__stdcall * Module32Next_type)(HANDLE, LPMODULEENTRY32);

  toolhelp_dll()
  :dll_base(toolhelp_dll_location::instance().data),
   CreateToolhelp32Snapshot((CreateToolhelp32Snapshot_type) load_proc("CreateToolhelp32Snapshot")),
   Module32First((Module32First_type) load_proc("Module32First")),
   Module32Next((Module32Next_type) load_proc("Module32Next"))
  {
  }

  CreateToolhelp32Snapshot_type CreateToolhelp32Snapshot;
  Module32First_type Module32First;
  Module32Next_type Module32Next;

  // Returns true if all required functions were found
  bool loaded() const
  {
    return (CreateToolhelp32Snapshot && Module32First && Module32Next);
  }
};

// This class represents psapi.dll and the functions that we use from it
struct psapi_dll: public dll_base
{
  typedef BOOL (__stdcall * EnumProcessModules_type)(HANDLE, HMODULE *, DWORD, LPDWORD);
  typedef BOOL (__stdcall * GetModuleInformation_type)(HANDLE, HMODULE, LPMODULEINFO, DWORD);

  psapi_dll()
  :dll_base(psapi_dll_location::instance().data),
   EnumProcessModules((EnumProcessModules_type) load_proc("EnumProcessModules")),
   GetModuleInformation((GetModuleInformation_type) load_proc("GetModuleInformation"))
  {
  }

  EnumProcessModules_type EnumProcessModules;
  GetModuleInformation_type GetModuleInformation;

  // Returns true if all required functions were found
  bool loaded() const
  {
    return (EnumProcessModules && GetModuleInformation);
  }
};

// Declaration of home-grown alternate implementations, in case some functions
//  are not found in the DLLs
BOOL __stdcall boost_SymFromAddr(HANDLE process, DWORD64 address, PDWORD64 displacement, PSYMBOL_INFO symbol);

// This class represents dbghelp.dll and the functions that we use from it
// It also keeps track of the dll's initialized state and will call Cleanup
//  if necessary before it unloads
class dbghelp_dll: public dll_base
{
  public:
    typedef BOOL (__stdcall * SymInitialize_type)(HANDLE, PSTR, BOOL);
    typedef BOOL (__stdcall * SymCleanup_type)(HANDLE);
    typedef DWORD (__stdcall * SymSetOptions_type)(DWORD);
    typedef PVOID (__stdcall * SymFunctionTableAccess_type)(HANDLE, DWORD64);
    typedef DWORD64 (__stdcall * SymGetModuleBase_type)(HANDLE, DWORD64);
    typedef BOOL (__stdcall * StackWalk_type)(DWORD, HANDLE, HANDLE,
        LPSTACKFRAME64, PVOID, PREAD_PROCESS_MEMORY_ROUTINE64,
        PFUNCTION_TABLE_ACCESS_ROUTINE64, PGET_MODULE_BASE_ROUTINE64,
        PTRANSLATE_ADDRESS_ROUTINE64);
    typedef DWORD64 (__stdcall * SymLoadModule_type)(HANDLE, HANDLE, PSTR, PSTR, DWORD64, DWORD);
    typedef PIMAGE_NT_HEADERS (__stdcall * ImageNtHeader_type)(PVOID);
    typedef PIMAGE_SECTION_HEADER (__stdcall * ImageRvaToSection_type)(PIMAGE_NT_HEADERS, PVOID, ULONG);
    typedef BOOL (__stdcall * SymGetModuleInfo_type)(HANDLE, DWORD64, PIMAGEHLP_MODULE64);
    typedef BOOL (__stdcall * SymGetLineFromAddr_type)(HANDLE, DWORD64, PDWORD, PIMAGEHLP_LINE64);
    typedef DWORD (__stdcall * UnDecorateSymbolName_type)(PCSTR, PSTR, DWORD, DWORD);
    typedef BOOL (__stdcall * SymGetSearchPath_type)(HANDLE, PSTR, DWORD);
    typedef BOOL (__stdcall * SymSetSearchPath_type)(HANDLE, PSTR);
    typedef BOOL (__stdcall * SymFromAddr_type)(HANDLE, DWORD64, PDWORD64, PSYMBOL_INFO);
    typedef BOOL (__stdcall * SymEnumSymbolsForAddr_type)(HANDLE, DWORD64,
        PSYM_ENUMERATESYMBOLS_CALLBACK, PVOID);
    typedef BOOL (__stdcall * SymGetSymFromAddr_type)(HANDLE, DWORD64, PDWORD64, PIMAGEHLP_SYMBOL64);

  private:
    // This is true if the DLL was found, loaded, all required functions found, and initalized without error
    bool loaded_;

    // Helper function for getting the full path and file name for a module
    // Returns std::string() on error
    static std::string get_module_file_name(const HMODULE module)
    {
      std::size_t size = 32;
      scoped_array<char> ret;
      do
      {
        size *= 2;
        ret.reset(new char[size]);
        std::memset(ret.get(), 0, size);
        if (!GetModuleFileName(module, ret.get(), size))
        {
          debug_Win32("GetModuleFileName");
          return std::string();
        }
      } while (ret[size - 1] != 0);
      return ret.get();
    }

    // Load all the modules in this process, using the PSAPI
    void load_modules(const psapi_dll & psapi)
    {
      DWORD bytes_needed;
      psapi.EnumProcessModules(GetCurrentProcess(), (HMODULE *) &bytes_needed, 0, &bytes_needed);
      SetLastError(0);
      const std::size_t modules_size = bytes_needed / sizeof(HMODULE);
      const scoped_array<HMODULE> modules(new HMODULE[modules_size]);
      if (!psapi.EnumProcessModules(GetCurrentProcess(), modules.get(), bytes_needed, &bytes_needed))
      {
        debug_Win32("EnumProcessModules");
        return;
      }

      for (std::size_t i = 0; i < modules_size; ++i)
      {
        MODULEINFO info;
        if (!psapi.GetModuleInformation(GetCurrentProcess(), modules[i], &info, sizeof(info)))
        {
          debug_Win32("GetModuleInformation (" + to_hex_string(modules[i]) + ")");
          continue;
        }

        if (!SymLoadModule(process(), 0, const_cast<char *>(get_module_file_name(modules[i]).c_str()),
            0, reinterpret_cast<DWORD64>(info.lpBaseOfDll), info.SizeOfImage))
          debug_Win32(std::string("SymLoadModule (") +  + ")");
      }
    }

    // Load all the modules in this process, using the ToolHelp API
    void load_modules(const toolhelp_dll & toolhelp)
    {
      const HANDLE snapshot = toolhelp.CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, GetCurrentProcessId());
      if (snapshot == INVALID_HANDLE_VALUE)
      {
        debug_Win32("CreateToolhelp32Snapshot");
        return;
      }
      file_handle snapshot_guard(snapshot);

      MODULEENTRY32 module;
      std::memset(&module, 0, sizeof(module));
      module.dwSize = sizeof(module);
      if (!toolhelp.Module32First(snapshot, &module))
      {
        debug_Win32("Module32First");
        return;
      }

      do
      {
        if (!SymLoadModule(process(), 0, module.szExePath, module.szModule,
            reinterpret_cast<DWORD64>(module.modBaseAddr), module.modBaseSize))
          debug_Win32(std::string("SymLoadModule (") + module.szModule + ")");
      } while (toolhelp.Module32Next(snapshot, &module));
      if (GetLastError() != ERROR_NO_MORE_FILES)
        debug_Win32("Module32Next");
    }

    // This function will try anything to enumerate and load all of this process' modules
    void load_modules()
    {
      // Try the ToolHelp API first
      const toolhelp_dll toolhelp;
      if (toolhelp.loaded())
      {
        load_modules(toolhelp);
        return;
      }

      // Try PSAPI if ToolHelp isn't present
      const psapi_dll psapi;
      if (psapi.loaded())
      {
        load_modules(psapi);
        return;
      }

      // One of the two above should work.  The only possible case where they wouldn't is
      //  if an NT user deleted psapi.dll...
    }

    // Retrieves the current search path, or std::string() on error
    std::string get_search_path()
    {
      std::size_t size = MAX_PATH;
      scoped_array<char> buffer;
      do
      {
        size *= 2;
        buffer.reset(new char[size]);
        std::memset(buffer.get(), 0, size);
        if (!SymGetSearchPath(process(), buffer.get(), size))
        {
          debug_Win32("SymGetSearchPath");
          return std::string();
        }
      } while (buffer[size - 1] != 0);
      return buffer.get();
    }

    void initialize()
    {
      // Initialize dbghelp
      if (!SymInitialize(process(), 0, FALSE))
      {
        debug_Win32("SymInitialize");
        return;
      }

      // dbghelp.dll has been Initialized
      loaded_ = true;

      // Set the search path; if the get/set fails, just leave it as default
      if (!symbol_search_path::instance().data.empty())
      {
        const std::string old_path = get_search_path();
        if (!old_path.empty())
        {
          const std::string path = symbol_search_path::instance().data + ";" + old_path;
          if (!SymSetSearchPath(process(), const_cast<char *>(path.c_str())))
            debug_Win32("SymSetSearchPath");
        }
      }

      // Enumerate and load all process modules
      load_modules();
    }

    void uninitialize()
    {
      if (loaded_)
        if (!SymCleanup(process()))
          debug_Win32("SymCleanup");
    }

  public:
    dbghelp_dll()
    :dll_base(dbghelp_dll_location::instance().data),
     loaded_(false),
     SymInitialize((SymInitialize_type) load_proc("SymInitialize")),
     SymCleanup((SymCleanup_type) load_proc("SymCleanup")),
     SymSetOptions((SymSetOptions_type) load_proc("SymSetOptions")),
     SymFunctionTableAccess((SymFunctionTableAccess_type) load_proc("SymFunctionTableAccess64")),
     SymGetModuleBase((SymGetModuleBase_type) load_proc("SymGetModuleBase64")),
     StackWalk((StackWalk_type) load_proc("StackWalk64")),
     SymLoadModule((SymLoadModule_type) load_proc("SymLoadModule64")),
     ImageNtHeader((ImageNtHeader_type) load_proc("ImageNtHeader")),
     ImageRvaToSection((ImageRvaToSection_type) load_proc("ImageRvaToSection")),
     SymGetModuleInfo((SymGetModuleInfo_type) load_proc("SymGetModuleInfo64")),
     SymGetLineFromAddr((SymGetLineFromAddr_type) load_proc("SymGetLineFromAddr64")),
     UnDecorateSymbolName((UnDecorateSymbolName_type) load_proc("UnDecorateSymbolName")),
     SymGetSearchPath((SymGetSearchPath_type) load_proc("SymGetSearchPath")),
     SymSetSearchPath((SymSetSearchPath_type) load_proc("SymSetSearchPath")),
     SymFromAddr((SymFromAddr_type) load_proc("SymFromAddr")),
     SymEnumSymbolsForAddr((SymEnumSymbolsForAddr_type) load_proc("SymEnumSymbolsForAddr"))
    {
      // Inject replacement SymFromAddr on the fly, if necessary
      if (!SymFromAddr)
      {
        if (!SymGetSymFromAddr)
          SymGetSymFromAddr = (SymGetSymFromAddr_type) load_proc("SymGetSymFromAddr64");
        if (SymGetSymFromAddr)
          SymFromAddr = &boost_SymFromAddr;
      }

      // If any critical functions (or the DLL itself) are missing, give up
      //  (SymEnumSymbolsForAddr is optional)
      if (!SymInitialize || !SymCleanup || !SymSetOptions ||
          !SymFunctionTableAccess || !SymGetModuleBase || !StackWalk ||
          !SymLoadModule || !ImageNtHeader || !ImageRvaToSection ||
          !SymGetModuleInfo || !SymGetLineFromAddr || !UnDecorateSymbolName ||
          !SymGetSearchPath || !SymSetSearchPath || !SymFromAddr)
        return;

      // Set the options
      SymSetOptions(SYMOPT_DEFERRED_LOADS | SYMOPT_LOAD_LINES);

      initialize();
    }

    ~dbghelp_dll()
    {
      uninitialize();
    }

    // Uninitializes and initializes dbghelp.dll (also reloads process module list)
    void reset()
    {
      uninitialize();
      initialize();
    }

    // The following are functions that we absolutely require in order to do stack traces
    SymInitialize_type SymInitialize;
    SymCleanup_type SymCleanup;
    SymSetOptions_type SymSetOptions;
    SymFunctionTableAccess_type SymFunctionTableAccess;
    SymGetModuleBase_type SymGetModuleBase;
    StackWalk_type StackWalk;

    // The following are functions that we may be able to do without, but they are always
    //  provided if the functions above are provided (so we assume they will be present)
    SymLoadModule_type SymLoadModule;
    ImageNtHeader_type ImageNtHeader;
    ImageRvaToSection_type ImageRvaToSection;
    SymGetModuleInfo_type SymGetModuleInfo;
    SymGetLineFromAddr_type SymGetLineFromAddr;
    UnDecorateSymbolName_type UnDecorateSymbolName;
    SymGetSearchPath_type SymGetSearchPath;
    SymSetSearchPath_type SymSetSearchPath;

    // The following are functions that we provide alternate implementations for, if they
    //  are not provided.
    SymFromAddr_type SymFromAddr;

    // The following are optional functions, that we do not provide alternate implementations for
    SymEnumSymbolsForAddr_type SymEnumSymbolsForAddr;

    // The following are functions that are normally not used, but they are used to provide
    //  alternate implementations for the optional functions above
    SymGetSymFromAddr_type SymGetSymFromAddr;

    // Returns true if all required functions were found and the init was OK
    bool loaded() const { return loaded_; }

    // The value used to initialize dbghelp.dll
    // We always use the process id here, so it will work on 95/98/Me
    static HANDLE process() { return reinterpret_cast<HANDLE>(GetCurrentProcessId()); }
};

typedef singleton<dbghelp_dll> dbghelp_singleton;

//
// DLL alternative implementation functions
//

// Home-grown SymFromAddr
// This function is required because 2000 has a dbghelp.dll without it.
// Also, the MS redistributable dbghelp.dll (as of 2002-06-28) does not have it, so
//  it will not be in the dbghelp.dll for 95/98/Me/NT.
BOOL __stdcall boost_SymFromAddr(HANDLE process, DWORD64 address, PDWORD64 displacement, PSYMBOL_INFO symbol)
{
  const dbghelp_dll & dbghelp = dbghelp_singleton::instance();
  scoped_array<char> info_(new char[sizeof(IMAGEHLP_SYMBOL64) + symbol->MaxNameLen]);
  IMAGEHLP_SYMBOL64 * const info = reinterpret_cast<IMAGEHLP_SYMBOL64 *>(info_.get());
  std::memset(info, 0, sizeof(IMAGEHLP_SYMBOL64) + symbol->MaxNameLen);
  info->SizeOfStruct = sizeof(IMAGEHLP_SYMBOL64);
  info->MaxNameLength = symbol->MaxNameLen;
  info->Address = address;
  if (!dbghelp.SymGetSymFromAddr(process, address, displacement, info))
  {
    debug_Win32("SymGetSymFromAddr (from within alternate SymFromAddr (" +
        to_hex_string(reinterpret_cast<void *>(address)) + ") )", GetLastError());
    return FALSE;
  }
  std::strcpy(symbol->Name, info->Name);
  return TRUE;
}

//
// Stack-walking-related helper functions
//

struct logical_address
{
  // The base file name of the loaded module
  std::string module;

  // The virtual address of the loaded module
  void * module_base;

  // The 'section' and 'offset' of a logical address will never be more than 32 bits, even on Win64
  unsigned section, offset;

  logical_address()
  :module_base(0), section(0), offset(0) { }
};

// Translates a virtual address to a logical address
// May return a partially-filled-out logical_address on error:
//   if (!module.empty()) -> module, module_base are correct
//   if (section != 0) -> module, module_base, section, offset are all correct
logical_address virtual_address_to_logical_address(const void * const address)
{
  const dbghelp_dll & dbghelp = dbghelp_singleton::instance();

  const DWORD64 virtual_address = reinterpret_cast<DWORD64>(address);
  logical_address ret;

  // Get the module information for the module that contains that address
  IMAGEHLP_MODULE64 module_info;
  std::memset(&module_info, 0, sizeof(module_info));
  module_info.SizeOfStruct = sizeof(module_info);
  if (!dbghelp.SymGetModuleInfo(dbghelp.process(), virtual_address, &module_info))
  {
    debug_Win32("SymGetModuleInfo (" + to_hex_string(address) + ")");
    return ret;
  }

  void * const module_base = reinterpret_cast<void *>(module_info.BaseOfImage);
  // A relative virtual address cannot be more than 32 bits, even on Win64
  const DWORD relative_virtual_address = static_cast<DWORD>(virtual_address - module_info.BaseOfImage);

  // We have successfully looked up the module, so save its name and base address
  ret.module = module_info.ModuleName;
  ret.module_base = module_base;

  // Get the NT header for this module
  IMAGE_NT_HEADERS * const nt_header = dbghelp.ImageNtHeader(module_base);
  if (nt_header == 0)
  {
    debug_Win32(std::string("ImageNtHeader (") + module_info.ImageName + ")");
    return ret;
  }

  // Find which section in the module this address refers to
  const IMAGE_SECTION_HEADER * section = dbghelp.ImageRvaToSection(nt_header,
      module_base, relative_virtual_address);
  if (section == 0)
  {
    debug_Win32("ImageRvaToSection");
    return ret;
  }

  ret.section = section - IMAGE_FIRST_SECTION(nt_header) + 1;
  ret.offset = relative_virtual_address - section->VirtualAddress;

  return ret;
}

// This is used as the callback that StackWalk uses to read process memory
// We pass it through to ReadProcessMemory, after making the call 64-bit safe,
//  and make a note of any errors returned.
BOOL CALLBACK DoReadProcessMemory(HANDLE, DWORD64 address,
    void * buffer, DWORD size, DWORD * bytes_read)
{
  // On Win32, only 32 bits of 'address' is used

  // We use a separate SIZE_T variable for the bytes_read result value,
  //  because it is 64 bits on Win64
  SIZE_T bytes_read_;
  const BOOL ret = ReadProcessMemory(GetCurrentProcess(),
      reinterpret_cast<const void *>(address), buffer, size, &bytes_read_);
  *bytes_read = static_cast<DWORD>(bytes_read_);

  if (!ret)
    debug_Win32("ReadProcessMemory (" + to_hex_string(address) + ")", GetLastError());
  return ret;
}

// Function that does the actual stack tracing of a thread, given its context
template <typename OutputIterator>
void trace_stack(CONTEXT & ctx,
    const HANDLE thread,
    OutputIterator out)
{
  const dbghelp_dll & dbghelp = dbghelp_singleton::instance();

  STACKFRAME64 frame;
  std::memset(&frame, 0, sizeof(frame));

#if defined(_M_IX86)
  const DWORD image_file_type = IMAGE_FILE_MACHINE_I386;
  frame.AddrPC.Mode = AddrModeFlat;
  frame.AddrPC.Offset = ctx.Eip;
  frame.AddrFrame.Mode = AddrModeFlat;
  frame.AddrFrame.Offset = ctx.Ebp;
#elif defined(_M_IA64)
  const DWORD image_file_type = IMAGE_FILE_MACHINE_IA64;
  frame.AddrPC.Mode = AddrModeFlat;
  frame.AddrPC = ctx.StIIP;
  frame.AddrBStore.Mode = AddrModeFlat;
  frame.AddrBStore = ctx.RsBSP;
#elif defined(_M_AMD64)
  const DWORD image_file_type = IMAGE_FILE_MACHINE_AMD64;
  frame.AddrPC.Mode = AddrModeFlat;
  frame.AddrPC = ctx.Rip;
  frame.AddrFrame.Mode = AddrModeFlat;
  frame.AddrFrame = ctx.Rbp;
#else
  #error Unknown machine type!
#endif

  while (dbghelp.StackWalk(image_file_type, dbghelp.process(), thread,
      &frame, &ctx, &DoReadProcessMemory, dbghelp.SymFunctionTableAccess,
      dbghelp.SymGetModuleBase, 0))
  {
    // Sometimes StackWalk will leave the last error set, which causes us to report an
    //  (incorrect) error after this loop.
    SetLastError(0);
    *out++ = reinterpret_cast<void *>(frame.AddrPC.Offset);
    if (frame.AddrReturn.Offset == 0)
      break;
  }
  if (GetLastError() != 0)
    debug_Win32("StackWalk (PC=" + to_hex_string(frame.AddrPC.Offset) +
        ", RET=" + to_hex_string(frame.AddrReturn.Offset) + ")");
}

// This is the structure passed to the slave thread, when that implementation is used.
template <typename OutputIterator>
struct thread_params
{
  HANDLE parent;
  OutputIterator * out;

  // This is the slave thread code itself.
  static DWORD WINAPI thread_proc(void * arg)
  {
    const dbghelp_dll & dbghelp = dbghelp_singleton::instance();

    // Read parameters into local variables
    thread_params * const params = reinterpret_cast<thread_params *>(arg);
    OutputIterator & out = *params->out;

    // Suspend the calling thread
    if (SuspendThread(params->parent) == DWORD(-1))
      details::stack::debug_Win32("SuspendThread");

    // Get its context
    CONTEXT ctx;
    std::memset(&ctx, 0, sizeof(ctx));
    ctx.ContextFlags = CONTEXT_FULL;
    if (!GetThreadContext(params->parent, &ctx))
      details::stack::debug_Win32("GetThreadContext");

    // Trace its stack
    details::stack::trace_stack(ctx, params->parent, out);

    // Resume it
    if (ResumeThread(params->parent) == DWORD(-1))
      details::stack::debug_Win32("ResumeThread");

    // And exit
    return 0;
  }
};

std::string unmangle_name(const char * const name)
{
  const dbghelp_dll & dbghelp = dbghelp_singleton::instance();
  std::size_t size = 32;
  std::size_t ret_size;
  scoped_array<char> buffer;
  do
  {
    size *= 2;
    buffer.reset(new char[size]);
    ret_size = dbghelp.UnDecorateSymbolName(name, buffer.get(), size, UNDNAME_COMPLETE);
    if (ret_size == 0)
    {
      debug_Win32("UnDecorateSymbolName (" + std::string(name) + ")");
      return name;
    }
  } while (ret_size == size - 1);
  return buffer.get();
}

BOOL CALLBACK enum_symbols_callback(PSYMBOL_INFO info, ULONG, PVOID ret_)
{
  std::string & ret = *reinterpret_cast<std::string *>(ret_);
  if (!ret.empty())
    ret += " / ";
  // If this is a MS mangled C++ symbol name, try to unmangle it
  if (info->Name[0] == '?')
    ret += details::stack::unmangle_name(info->Name);
  else
    ret += info->Name;
  return TRUE;
}

} // namespace stack
} // namespace details

// These are tags that can be passed to 'walk_stack' that control which implementation
//  is used.  These are a Win32 extension.
struct stack_Win32_use_exception { };
struct stack_Win32_use_thread { };
struct stack_Win32_use_direct { };

namespace details {
namespace stack {

// Exception implementation
template <typename OutputIterator>
void get_context_and_walk_stack(const OutputIterator & out, stack_Win32_use_exception)
{
  // Get the context record implicitly from raising an exception
  __try {
    RaiseException(13, 0, 0, 0);
  }
  __except (trace_stack(*GetExceptionInformation()->ContextRecord,
      GetCurrentThread(), out), EXCEPTION_EXECUTE_HANDLER)
  {
  }
  // Note: we have to walk the stack in the filter function, because it's already changed
  //  by the time the exception is handled (or continued).
  // This may cause problems because we're calling a user-specified function (*out++) from
  //  within a filter expression.  
}

// Thread implementation
template <typename OutputIterator>
void get_context_and_walk_stack(OutputIterator out, stack_Win32_use_thread)
{
  // Get the context record by spawning off a worker thread

  // Set up thread parameters
  thread_params<OutputIterator> params;
  if (!DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), GetCurrentProcess(), &params.parent, 0, FALSE,
      DUPLICATE_SAME_ACCESS))
    debug_Win32("DuplicateHandle");
  file_handle parent_guard(params.parent);
  params.out = &out;

  // Create the thread
  DWORD child_id;
  const HANDLE child = CreateThread(0, 0, &thread_params<OutputIterator>::thread_proc, &params, 0, &child_id);
  if (child == 0)
    debug_Win32("CreateThread");
  file_handle child_guard(child);

  // The child thread will suspend this thread sometime between CreateThread and
  //  WaitForSingleObject, walk this thread's stack, resume this thread, and exit.

  // Wait for the child thread to complete
  if (WaitForSingleObject(child, INFINITE) != WAIT_OBJECT_0)
    debug_Win32("WaitForSingleObject");
}

template <typename OutputIterator>
void get_context_and_walk_stack(const OutputIterator & out, stack_Win32_use_direct)
{
  // Get the context record directly

  // You're really not supposed to do this...
  //  It does not work on 98
  CONTEXT ctx;
  std::memset(&ctx, 0, sizeof(ctx));
  ctx.ContextFlags = CONTEXT_FULL;
  if (!GetThreadContext(GetCurrentThread(), &ctx))
    debug_Win32("GetThreadContext");

  trace_stack(ctx, GetCurrentThread(), out);
}

} // namespace stack
} // namespace details

// Walk the stack, using the given implementation ('Tag')
// This is a Win32 extension
template <typename OutputIterator, typename Tag>
void walk_stack(const OutputIterator & out, Tag)
{
  // Acquire the thread lock
  details::stack::critical_section::scoped_lock lock(details::stack::dbghelp_lock::instance());

  // Ensure dbghelp.dll is loaded
  details::stack::dbghelp_dll & dbghelp = details::stack::dbghelp_singleton::instance();

  // If there's no dbghelp.dll, just give up
  if (!dbghelp.loaded())
    return;

  details::stack::get_context_and_walk_stack(out, Tag());
}

// Walk the stack, using the default implementation
template <typename OutputIterator>
void walk_stack(const OutputIterator & out)
{ walk_stack(out, stack_Win32_use_exception()); }

template <typename InputIterator, typename OutputIterator>
void stack_symbols(InputIterator first, const InputIterator & last, OutputIterator out)
{
  // Acquire the thread lock
  details::stack::critical_section::scoped_lock lock(details::stack::dbghelp_lock::instance());

  // Ensure dbghelp.dll is loaded
  details::stack::dbghelp_dll & dbghelp = details::stack::dbghelp_singleton::instance();

  while (first != last)
  {
    const void * const p = *first++;

    // If there's no dbghelp.dll, just give up
    if (!dbghelp.loaded())
    {
      *out++ = details::stack::to_hex_string(p);
      continue;
    }

    std::string ret;

    // Output the virtual address

    ret += details::stack::to_hex_string(p);
    ret += "\t";

    // Output the logical address

    details::stack::logical_address tmp = details::stack::virtual_address_to_logical_address(p);
    if (tmp.module.empty())
      ret += "Unknown";
    else
      ret += tmp.module;
    ret += " (";
    ret += details::stack::to_hex_string(tmp.module_base);
    ret += ") ";
    ret += lexical_cast<std::string>(tmp.section);
    ret += ":";
    ret += details::stack::to_hex_string(tmp.offset);
    ret += "\t";

    // Output the source info

    IMAGEHLP_LINE64 line;
    std::memset(&line, 0, sizeof(line));
    line.SizeOfStruct = sizeof(line);
    DWORD displacement;
    if (!dbghelp.SymGetLineFromAddr(dbghelp.process(), reinterpret_cast<DWORD64>(p),
        &displacement, &line))
      details::stack::debug_Win32("SymGetLineFromAddr (" + details::stack::to_hex_string(p) + ")");
    else
    {
      ret += line.FileName;
      ret += ":";
      ret += lexical_cast<std::string>(line.LineNumber);
    }
    ret += "\t";

    // Output the symbol name(s)

    if (dbghelp.SymEnumSymbolsForAddr)
    {
      std::string name;
      if (!dbghelp.SymEnumSymbolsForAddr(dbghelp.process(), reinterpret_cast<DWORD64>(p),
          &details::stack::enum_symbols_callback, &name))
        details::stack::debug_Win32("SymEnumSymbolsForAddr (" + details::stack::to_hex_string(p) + ")");
      else
        ret += name;
    }
    else
    {
      char info_[sizeof(SYMBOL_INFO) + MAX_PATH];
      SYMBOL_INFO * const info = reinterpret_cast<SYMBOL_INFO *>(info_);
      std::memset(info, 0, sizeof(SYMBOL_INFO) + MAX_PATH);
      info->SizeOfStruct = sizeof(SYMBOL_INFO);
      info->MaxNameLen = MAX_PATH;
      DWORD64 offset;
      if (dbghelp.SymFromAddr(dbghelp.process(), reinterpret_cast<DWORD64>(p), &offset, info))
      {
        // If this is a MS mangled C++ symbol name, try to unmangle it
        if (info->Name[0] == '?')
          ret += details::stack::unmangle_name(info->Name);
        else
          ret += info->Name;
      }
      else
        details::stack::debug_Win32("SymFromAddr (" + details::stack::to_hex_string(p) + ")");
    }

    *out++ = ret;
  }
}

// These functions allow the setting of different parameters used by Boost.StackWalk
// They must be called before any other Boost.StackWalk functions

void stack_Win32_set_dbghelp_dll_location(const std::string & location)
{ details::stack::dbghelp_dll_location::instance().data = location; }

void stack_Win32_set_toolhelp_dll_location(const std::string & location)
{ details::stack::toolhelp_dll_location::instance().data = location; }

void stack_Win32_set_psapi_dll_location(const std::string & location)
{ details::stack::psapi_dll_location::instance().data = location; }

void stack_Win32_set_symbol_search_path(const std::string & path)
{ details::stack::symbol_search_path::instance().data = path; }

// This function reloads the module list kept in dbghelp.dll
// Call it after loading or unloading DLLs
// This is a Win32 extension

void stack_Win32_refresh_module_list()
{
  // Acquire the thread lock
  details::stack::critical_section::scoped_lock lock(details::stack::dbghelp_lock::instance());

  // Ensure dbghelp.dll is loaded
  details::stack::dbghelp_dll & dbghelp = details::stack::dbghelp_singleton::instance();

  // If there's no dbghelp.dll, just give up
  if (!dbghelp.loaded())
    return;

  // Refresh the module list for dbghelp.dll
  dbghelp.reset();
}

} // namespace boost

#endif